__version__ = "1.0.3"
__description__ = "Jumpstarter Console"
__long_description__ = """
Jumpstarter Console - A command line tool for managing your Jumpstarter projects.
please visit https://github.com/jumpstarter-io/jsc for more information
"""
